/***************************************************************************
                          cmudview.h
                      -------------------
		description					 : class providing a colored output view
    begin                : Wed Sep 8 1999
    copyright            : (C) 1999 by Stephan Uhlmann                         
    email                : suhlmann@gmx.de                                     
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   * 
 *                                                                         *
 ***************************************************************************/


#ifndef CMUDVIEW_H
#define CMUDVIEW_H

#include <qtableview.h>
#include <qlist.h>
#include <qpainter.h>
#include <qkeycode.h>
#include "kmuddoc.h"

/** A chunk of text with defined color/font settings */
struct MudViewChunkData
{
	QString s;
  QColor fgc;
  QColor bgc;
  QFont font;
};

/** Struct handling one line of text consisting of a list of MudViewChunkData's */
struct MudViewLineData
{
	QList<MudViewChunkData> chunks;
  int lineHeight;
};


/**
	* Class providing a colored output view.
	* Interprets ansi/vt100 escape sequences.
  *@author Stephan Uhlmann
  */

class CMudView : public QTableView  {

	Q_OBJECT

public: 
	/** constructor */
	CMudView(KmudDoc* document, QWidget * parent=0, const char * name=0);
	/** destructor */
	~CMudView();


	/** returns the length of a line in pixels */
	int lineLength(int line);
	/** returns the total number of lines */
	int numLines();
	/** appends a string with a specified color/background/font to the view.
			automagically interprets ansi/vt100 escape sequences and adapts the color/font settings */
  void append(QString s, QColor fgc, QColor bgc, QFont font);

	/** sets the default foreground color to use */
	void setForeground(QColor fgc);
	/** sets the default background color to use */
	void setBackground(QColor bgc);
	/** sets the default font to use */
	void setFont(QFont font);

	/** Sets the interpretation type */
	void setInterpretType(int i);
	
	// sets/gets debug highlight color
	void setDebugHighlightColor(QColor color);
	QColor getDebugHighlightColor();
	
	/** sets the colors for ANSI */
	void setOutputANSIBlack(QColor color);
	void setOutputANSIDarkRed(QColor color);
	void setOutputANSIDarkGreen(QColor color);
	void setOutputANSIDarkYellow(QColor color);
	void setOutputANSIDarkBlue(QColor color);
	void setOutputANSIDarkMagenta(QColor color);
	void setOutputANSIDarkCyan(QColor color);
	void setOutputANSIDarkWhite(QColor color);
	
	void setOutputANSIGray(QColor color);
	void setOutputANSILightRed(QColor color);
	void setOutputANSILightGreen(QColor color);
	void setOutputANSILightYellow(QColor color);
	void setOutputANSILightBlue(QColor color);
	void setOutputANSILightMagenta(QColor color);
	void setOutputANSILightCyan(QColor color);
	void setOutputANSILightWhite(QColor color);

	/** sets background colors */
	void setOutputBgBlack(QColor color);
	void setOutputBgRed(QColor color);
	void setOutputBgGreen(QColor color);
	void setOutputBgYellow(QColor color);
	void setOutputBgBlue(QColor color);
	void setOutputBgMagenta(QColor color);
	void setOutputBgCyan(QColor color);
	void setOutputBgWhite(QColor color);
		
	/** gets the default foreground color to use */
	QColor getForeground();
	/** gets the default background color to use */
	QColor getBackground();
	/** gets the default font to use */
	QFont getFont();

	int getInterpretType();
	
	/** sets the colors for ANSI */
	QColor getOutputANSIBlack();
	QColor getOutputANSIDarkRed();
	QColor getOutputANSIDarkGreen();
	QColor getOutputANSIDarkYellow();
	QColor getOutputANSIDarkBlue();
	QColor getOutputANSIDarkMagenta();
	QColor getOutputANSIDarkCyan();
	QColor getOutputANSIDarkWhite();
	
	QColor getOutputANSIGray();
	QColor getOutputANSILightRed();
	QColor getOutputANSILightGreen();
	QColor getOutputANSILightYellow();
	QColor getOutputANSILightBlue();
	QColor getOutputANSILightMagenta();
	QColor getOutputANSILightCyan();
	QColor getOutputANSILightWhite();
	
	/** gets colors for background */
	QColor getOutputBgBlack();
	QColor getOutputBgRed();
	QColor getOutputBgGreen();
	QColor getOutputBgYellow();
	QColor getOutputBgBlue();
	QColor getOutputBgMagenta();
	QColor getOutputBgCyan();
	QColor getOutputBgWhite();
	
	/** sets maximum number of lines in the view */
	void setMaxLines(int);

	/** sets useBeep -- use \007 */
	void setUseBeep(bool);
	/** gets useBeep -- use \007 */
	bool getUseBeep();
	
signals:
	/** signal emitted when text in the widget chaged */
	void textChanged();

protected:
	/** paints the cell (line) in the specified row/column
		* using the given QPainter
		*/
	void paintCell(QPainter*, int row, int col );
	/** return the height of the specified row in pixels */
  int cellHeight(int row);
	/** redefined resizeEvent to set the cellWidth to the total width */
  virtual void resizeEvent(QResizeEvent *);
  /** overwrites the mousePressEvent-Method from QWidget
implements the start of marking a text-area */
  virtual void mousePressEvent (QMouseEvent* e);
  /** overwrites the mouseMoveEvent from QWidget
implements the marking of a area */
  virtual void mouseMoveEvent (QMouseEvent* e);


private:
	/** initializes the view, called in constructor and when clearscreen
			escape sequence was received */
	void init();
  /** copies the marked text in the X-Clipboard */
  void copyMarkedText ();
  /** gets the number of the character in the line row and on the position xPos */
  int getCharacterNumber (int Row, int xPos);
  /** returns the text in the marked area */
  QString getMarkedText ();
  /** returns true, if the specified row is marked */
  int rowMarked (int Row);

	QList< MudViewLineData > data;

	QColor defaultForeground;
	QColor defaultBackground;
	QFont defaultFont;

	int interpretType;

	QColor DEBUG_HIGHLIGHT;
	
	QColor BLACK;
	QColor DARK_RED;
	QColor DARK_GREEN;
	QColor DARK_YELLOW;
	QColor DARK_BLUE;
	QColor DARK_MAGENTA;
	QColor DARK_CYAN;
	QColor DARK_WHITE;
	
	QColor GRAY;
	QColor LIGHT_RED;
	QColor LIGHT_GREEN;
	QColor LIGHT_YELLOW;
	QColor LIGHT_BLUE;
	QColor LIGHT_MAGENTA;
	QColor LIGHT_CYAN;
	QColor LIGHT_WHITE;
	
	QColor BG_BLACK;
	QColor BG_RED;
	QColor BG_GREEN;
	QColor BG_YELLOW;
	QColor BG_BLUE;
	QColor BG_MAGENTA;
	QColor BG_CYAN;
	QColor BG_WHITE;
	
	QColor actualForeground;
	QColor actualBackground;
	QFont actualFont;

	int maxLines;
  /** starting column of the marked area */
  int markedAreaStartCol;
  /** column of the ending of the marked area */
  int markedAreaEndCol;
  /** row of the end of the marked area */
  int markedAreaEndRow;
  /** starting row of the marked area */
  int markedAreaStartRow;
	int lineX;

  bool useDebugHighlight;

  /** the Document */
  KmudDoc* doc;

  /** To take advantage of the \007 char */
  bool useBeep;
  /** is set, if some text is marked */
  bool markedText;
    	
public slots:
  /** Scrolls around when certain keys are pressed in CMudLineEdit. */
  void slotScrollKey(QKeyEvent *);
};

#endif


























